package com.me.asteroids;

import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.math.Vector2;
import com.me.asteroids.components.AccelerationComponent;
import com.me.asteroids.components.AsteroidComponent;
import com.me.asteroids.components.BulletComponent;
import com.me.asteroids.components.ColliderComponent;
import com.me.asteroids.components.DebrisComponent;
import com.me.asteroids.components.DecayComponent;
import com.me.asteroids.components.ModelComponent;
import com.me.asteroids.components.PlayerComponent;
import com.me.asteroids.components.PositionComponent;
import com.me.asteroids.components.VelocityComponent;
import com.me.asteroids.components.model.LineModel;
import com.me.asteroids.components.model.Model;
import com.me.asteroids.components.model.PolygonModel;
import com.me.common.ecs.Engine;
import com.me.common.ecs.Entity;

import java.util.Arrays;

import static com.me.asteroids.Constants.rand;

public class EntityFactory {

    private static final Vector2 tmp = new Vector2();
    private static final Vector2 tmp2 = new Vector2();
    private static final Vector2 tmpA = new Vector2();
    private static final Vector2 tmpB = new Vector2();

    private static final ColliderComponent COLLIDER = new ColliderComponent();

    private static Entity createEntity(Engine engine) {
        Entity entity = engine.createEntity();
        if (Constants.DEBUG) {
            System.out.println("Spawned entity: " + entity);
        }
        return entity;
    }

    public static Entity createPlayer(Engine engine) {
        PositionComponent position = new PositionComponent();
        position.position = new Vector2(Constants.HALF_WIDTH, Constants.HALF_HEIGHT);
        position.rotation = 90;

        VelocityComponent velocity = new VelocityComponent();
        velocity.velocity = new Vector2(0f, 0f);
        velocity.maxVelocity = 10f;

        ModelComponent model = new ModelComponent();
        model.model = new PolygonModel(Color.WHITE);
        model.model.setVertices(new float[]{
                0f, 0.5f,         // tip
                -5/16f, -0.5f,     // bottom left
                -1/8f, -5/16f,     // indent
                1/8f, -5/16f,      // indent
                5/16f, -0.5f,      // bottom right
        });

        AccelerationComponent accel = new AccelerationComponent();
        accel.acceleration = new Vector2(0, 1f);

        ModelComponent afterBurnerModel = new ModelComponent();
        afterBurnerModel.model = new LineModel(Color.BLUE);
        afterBurnerModel.model.setVertices(new float[] {
                -2/16f, -5/16f,
                0f, -0.8f,
                2/16f, -5/16f
        });

        PositionComponent afterBurnerPosition = new PositionComponent();
        afterBurnerPosition.position = new Vector2(Constants.HALF_WIDTH, Constants.HALF_HEIGHT);
        afterBurnerPosition.rotation = 90;

        Entity player = createEntity(engine);
        Entity afterBurner = createEntity(engine);
        afterBurner.addComponent(afterBurnerPosition);
        afterBurner.addComponent(velocity);
        afterBurner.addComponent(afterBurnerModel);

        player.addComponent(position);
        player.addComponent(velocity);
        player.addComponent(model);
        player.addComponent(accel);
        player.addComponent(COLLIDER);
        player.addComponent(new PlayerComponent(afterBurner));

        return player;
    }


    private static LineModel[] getLineModels(Model playerModel) {
        float[] vertices = playerModel.getVertices();

        LineModel[] models = new LineModel[(vertices.length / 2)];
        for (int aIndex = 0, n = vertices.length; aIndex < n; aIndex += 2) {
            int bIndex = (aIndex + 2 == n) ? 0 : (aIndex + 2);

            tmpA.set(vertices[aIndex], vertices[aIndex + 1]);
            tmpB.set(vertices[bIndex], vertices[bIndex + 1]);

            tmp.set(tmpB).sub(tmpA).scl(0.5f);
            tmp2.set(tmpA).add(tmp);                // tmp2 = center of line segment

            tmpA.sub(tmp2);                         // make tmpA, tmpB relative to line center
            tmpB.sub(tmp2);

            LineModel model = new LineModel(playerModel.getColor());
            model.setVertices(new float[]{
                    tmpA.x, tmpA.y,
                    tmpB.x, tmpB.y,
            });
            model.setPosition(tmp2);                // Position = line center

            models[aIndex / 2] = model;
        }
        return models;
    }

    public static Entity[] createDebris(Engine engine, Entity entity) {
        Vector2 playerVelocity = entity.getComponent(VelocityComponent.class).velocity;
        PositionComponent playerPosition = entity.getComponent(PositionComponent.class);
        LineModel[] models = getLineModels(entity.getComponent(ModelComponent.class).model);
        Vector2 explosionCenter = tmp.set(playerPosition.position).sub(Utils.setUnitVectorAngle(tmp2, playerPosition.rotation).scl(0.125f));

        Entity[] entities = new Entity[models.length];
        for (int i = 0, n = models.length; i < n; i++) {
            ModelComponent model = new ModelComponent();
            model.model = models[i];

            PositionComponent position = new PositionComponent();
            position.position = new Vector2(model.model.getPosition());
            position.rotation = 90;

            VelocityComponent velocity = new VelocityComponent();
            velocity.velocity = new Vector2(models[i].getPosition())
                    .sub(explosionCenter)
                    .nor()                                      // Direction from explosion center to center of piece
                    .rotate(rand.nextFloat(-15, 15))            // Slightly alter the direction each piece flies off in
                    .scl(rand.nextFloat(2f, 2.5f))               // Give each piece a slightly different speed
                    .add(tmp2.set(playerVelocity).scl(0.75f));  // Maintain 75% of the player's velocity at impact
            velocity.angularVelocity = rand.nextFloat(-60, 60); // Make each piece spin at a different rate

            Entity debris = createEntity(engine);
            debris.addComponent(position);
            debris.addComponent(velocity);
            debris.addComponent(model);
            debris.addComponent(new DecayComponent(rand.nextFloat(0.5f, 2.5f)));
            debris.addComponent(new DebrisComponent());
            entities[i] = debris;
        }
        return entities;
    }


    public static Entity createBullet(Engine engine, Entity player) {
        float[] modelVertices = player.getComponent(ModelComponent.class).model.getVertices();
        float rotation = player.getComponent(PositionComponent.class).rotation;

        Vector2 direction = Utils.setUnitVectorAngle(tmp, rotation);

        VelocityComponent velocity = new VelocityComponent();
        velocity.velocity = new Vector2(direction).scl(12.5f);

        PositionComponent position = new PositionComponent();
        position.position = new Vector2(modelVertices[0], modelVertices[1]);
        position.rotation = rotation;

        ModelComponent model = new ModelComponent();
        model.model = new PolygonModel(Color.YELLOW);
        model.model.setVertices(new float[]{
                1/40f, 0f,
                -1/40f, 0f,
                -1/40f, -4/40f,
                1/40f, -4/40f,
        });
        model.model.setRotation(position.rotation);
        model.model.setPosition(position.position);


        Entity bullet = createEntity(engine);
        bullet.addComponent(position);
        bullet.addComponent(velocity);
        bullet.addComponent(model);
        bullet.addComponent(COLLIDER);
        bullet.addComponent(new BulletComponent());
        return bullet;
    }

    public static Entity createAsteroid(Engine engine) {
        // Creates an asteroid entity with position and velocity unset - the AsteroidSpawningSystem
        // is responsible for setting its position and velocity
        PositionComponent position = new PositionComponent();
        VelocityComponent velocity = new VelocityComponent();

        ModelComponent model = new ModelComponent();
        model.model = new PolygonModel(Color.WHITE);
        float size = rand.nextFloat(1f, 1.75f);
        model.model.setVertices(new AsteroidFactory()
                .setVertexCount(32)
                .setSize(size)
                .setSizeVariation(size * 0.5f)
                .sizeRelativeToLast()
                .generate());

        Entity asteroid = createEntity(engine);
        asteroid.addComponent(position);
        asteroid.addComponent(velocity);
        asteroid.addComponent(model);
        asteroid.addComponent(COLLIDER);
        asteroid.addComponent(new AsteroidComponent());
        return asteroid;
    }

    private static float[] scaleAndRelativizeVertices(Vector2 position, float[] vertices, float scale) {
        float[] ret = new float[vertices.length];
        for (int j = 0, n = vertices.length; j < n; j += 2) {
            ret[j] = (vertices[j] - position.x) * scale;
            ret[j + 1] = (vertices[j + 1] - position.y) * scale;
        }
        return ret;
    }

    public static Entity[] splitAsteroidIntoChunks(Engine engine, Entity asteroid, int chunkCount, float chunkScale) {
        Vector2 asteroidVelocity = asteroid.getComponent(VelocityComponent.class).velocity;
        Model asteroidModel = asteroid.getComponent(ModelComponent.class).model;
        Vector2 asteroidPosition = asteroidModel.getPosition();
        float[] scaledVertices = scaleAndRelativizeVertices(asteroidPosition, asteroidModel.getVertices(), chunkScale);

        float angle = rand.nextFloat() * MathUtils.PI2;
        float angleStep = MathUtils.PI2 / chunkCount;
        Utils.setUnitVectorAngleRad(tmp, angle);

        Entity[] entities = new Entity[chunkCount];
        for (int i = 0; i < chunkCount; i++) {
            Vector2 chunkPosition = tmp2.set(asteroidPosition).add(tmp.scl(0.5f));

            ModelComponent model = new ModelComponent();
            model.model = new PolygonModel(asteroidModel.getColor());
            model.model.setVertices(Arrays.copyOf(scaledVertices, scaledVertices.length));
            model.model.setPosition(chunkPosition);

            PositionComponent position = new PositionComponent();
            position.position = new Vector2(chunkPosition);
            position.rotation = 90;

            VelocityComponent velocity = new VelocityComponent();
            velocity.velocity = new Vector2(tmp).nor().rotate(rand.nextFloat(-45, 45)).scl(asteroidVelocity.len() * 1.10f);
            velocity.angularVelocity = rand.nextFloat(-30, 30);

            AsteroidComponent asteroidComponent = new AsteroidComponent();
            asteroidComponent.generation = asteroid.getComponent(AsteroidComponent.class).generation + 1;

            angle += angleStep;
            Utils.setUnitVectorAngleRad(tmp, angle);

            Entity split = createEntity(engine);
            split.addComponent(model);
            split.addComponent(position);
            split.addComponent(velocity);
            split.addComponent(asteroidComponent);
            split.addComponent(COLLIDER);
            entities[i] = split;
        }

        return entities;
    }

}
